import os
from random import random
from tempfile import mkdtemp

import biocframe
import dolomite_mae
import dolomite_sce
import dolomite_se
import numpy
import pandas as pd
from dolomite_base import read_object, save_object
from genomicranges import GenomicRanges
from multiassayexperiment import MultiAssayExperiment
from singlecellexperiment import SingleCellExperiment
from summarizedexperiment import SummarizedExperiment


def test_stage_mae_basic():
    x = numpy.random.rand(1000, 200)
    x2 = (numpy.random.rand(1000, 200) * 10).astype(numpy.int32)

    sce = SingleCellExperiment(
        {"logcounts": x, "counts": x2},
        main_experiment_name="aaron's secret modality",
        row_data=biocframe.BiocFrame(
            {"foo": numpy.random.rand(1000), "bar": numpy.random.rand(1000)},
            row_names=["gene_sce_" + str(i) for i in range(1000)],
        ),
        column_data=biocframe.BiocFrame(
            {"whee": numpy.random.rand(200), "stuff": numpy.random.rand(200)},
            row_names=["cell_sce" + str(i) for i in range(200)],
        ),
    )

    se = SummarizedExperiment(
        {"counts": numpy.random.rand(100, 200)},
        row_data=biocframe.BiocFrame(
            {"foo": numpy.random.rand(100), "bar": numpy.random.rand(100)},
            row_names=["gene_se_" + str(i) for i in range(100)],
        ),
        column_data=biocframe.BiocFrame(
            {"whee": numpy.random.rand(200), "stuff": numpy.random.rand(200)},
            row_names=["cell_se" + str(i) for i in range(200)],
        ),
    )

    mae = MultiAssayExperiment(experiments={"jay_expt": sce, "aarons_expt": se})

    dir = os.path.join(mkdtemp(), "mae_simple")
    save_object(mae, dir)

    roundtrip = read_object(dir)
    assert isinstance(roundtrip, MultiAssayExperiment)
    assert roundtrip.experiment("jay_expt").shape == sce.shape
    assert (
        roundtrip.experiment("aarons_expt").shape == mae.experiment("aarons_expt").shape
    )
    assert len(mae.get_column_data()) == 2
    assert len(mae.get_sample_map()) == 400
    assert list(mae.get_column_data().get_row_names()) == [
        "unknown_sample_jay_expt",
        "unknown_sample_aarons_expt",
    ]


def test_stage_mae_complex():
    nrows = 200
    ncols = 6
    counts = numpy.random.rand(nrows, ncols)
    df_gr = pd.DataFrame(
        {
            "seqnames": [
                "chr1",
                "chr2",
                "chr2",
                "chr2",
                "chr1",
                "chr1",
                "chr3",
                "chr3",
                "chr3",
                "chr3",
            ]
            * 20,
            "starts": range(100, 300),
            "ends": range(110, 310),
            "strand": ["-", "+", "+", "*", "*", "+", "+", "+", "-", "-"] * 20,
            "score": range(0, 200),
            "GC": [random() for _ in range(10)] * 20,
        }
    )

    gr = GenomicRanges.from_pandas(df_gr)

    column_data_sce = pd.DataFrame(
        {
            "treatment": ["ChIP", "Input"] * 3,
        },
        index=["sce"] * 6,
    )
    column_data_se = pd.DataFrame(
        {
            "treatment": ["ChIP", "Input"] * 3,
        },
        index=["se"] * 6,
    )

    sample_map = pd.DataFrame(
        {
            "assay": ["sce", "se"] * 6,
            "primary": ["sample1", "sample2"] * 6,
            "colname": ["sce", "se"] * 6,
        }
    )

    sample_data = pd.DataFrame(
        {"samples": ["sample1", "sample2"]}, index=["sample1", "sample2"]
    )

    tsce = SingleCellExperiment(
        assays={"counts": counts},
        row_data=df_gr,
        column_data=column_data_sce,
        row_ranges=gr,
    )

    tse2 = SummarizedExperiment(
        assays={"counts": counts.copy()},
        row_data=df_gr.copy(),
        column_data=column_data_se.copy(),
    )

    mae = MultiAssayExperiment(
        experiments={"sce": tsce, "se": tse2},
        column_data=sample_data,
        sample_map=sample_map,
        metadata={"could be": "anything"},
    )

    dir = os.path.join(mkdtemp(), "mae_simple")
    save_object(mae, dir)

    roundtrip = read_object(dir)
    assert isinstance(roundtrip, MultiAssayExperiment)
    assert roundtrip.experiment("sce").shape == mae.experiment("sce").shape
    assert roundtrip.experiment("se").shape == mae.experiment("se").shape
    assert len(mae.get_column_data()) == len(sample_data)
    assert len(mae.get_sample_map()) == len(sample_map)
    assert list(mae.get_column_data().get_row_names()) == [
        "sample1",
        "sample2",
    ]



import json
import os

import biocutils
import dolomite_base as dl
import h5py
from multiassayexperiment import MultiAssayExperiment


@dl.save_object.register
@dl.validate_saves
def save_multi_assay_experiment(
    x: MultiAssayExperiment,
    path: str,
    data_frame_args: dict = None,
    assay_args: dict = None,
    **kwargs,
):
    """Method for saving
    :py:class:`~multiassayexperiment.MultiAssayExperiment.MultiAssayExperiment`
    objects to their corresponding file representations, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x:
            Object to be staged.

        path:
            Path to a directory in which to save ``x``.

        data_frame_args:
            Further arguments to pass to the ``save_object`` method for the
            row/column data.

        assay_args:
            Further arguments to pass to the ``save_object`` method for the
            assays.

        kwargs: Further arguments, ignored.

    Returns:
        ``x`` is saved to path.
    """
    os.mkdir(path)

    if data_frame_args is None:
        data_frame_args = {}

    if assay_args is None:
        assay_args = {}

    with open(os.path.join(path, "OBJECT"), "w", encoding="utf-8") as handle:
        handle.write(
            '{ "type": "multi_sample_dataset", "multi_sample_dataset": { "version": "1.0" } }'
        )

    # sample/column data
    _sample_path = os.path.join(path, "sample_data")
    dl.save_object(x.get_column_data(), _sample_path, **data_frame_args)

    # save alt expts.
    _expt_names = x.get_experiment_names()
    if len(_expt_names) > 0:
        _expt_path = os.path.join(path, "experiments")
        os.mkdir(_expt_path)

        with open(os.path.join(_expt_path, "names.json"), "w") as handle:
            json.dump(_expt_names, handle)

        for _aidx, _aname in enumerate(_expt_names):
            _expt_save_path = os.path.join(_expt_path, str(_aidx))
            try:
                dl.save_object(
                    x.experiment(_aname),
                    path=_expt_save_path,
                    data_frame_args=data_frame_args,
                    assay_args=assay_args,
                )
            except Exception as ex:
                raise RuntimeError(
                    "failed to stage experiment '"
                    + _aname
                    + "' for "
                    + str(type(x))
                    + "; "
                    + str(ex)
                )
        with h5py.File(os.path.join(path, "sample_map.h5"), "w") as handle:
            ghandle = handle.create_group("multi_sample_dataset")

            _sample_map = x.get_sample_map()
            for _aidx, _aname in enumerate(_expt_names):
                _indices_to_keep = [
                    idx
                    for idx, x in enumerate(_sample_map.get_column("assay"))
                    if x == _aname
                ]

                _colnames = biocutils.subset_sequence(
                    _sample_map.get_column("colname"), _indices_to_keep
                )
                _sample = biocutils.subset_sequence(
                    _sample_map.get_column("primary"), _indices_to_keep
                )

                i = biocutils.match(_sample, x.get_column_data().get_row_names())

                if (i == -1).any():
                    raise RuntimeError(
                        "Samples in 'sample_map' not presented in 'column_data' for ",
                        f"{_aname}.",
                    )

                j = biocutils.match(x.experiment(_aname).get_column_names(), _colnames)
                if (j == -1).any():
                    raise RuntimeError(
                        f"Column names in experiment '{_aname}' not presented in 'sample_map'."
                    )

                reorder = i[j.tolist()]

                dl.write_integer_vector_to_hdf5(
                    ghandle, name=str(_aidx), h5type="u4", x=reorder
                )

    _meta = x.get_metadata()
    if _meta is not None and len(_meta) > 0:
        dl.save_object(_meta, path=os.path.join(path, "other_data"))

    return



import json
import os

import dolomite_base as dl
import h5py
from biocframe import BiocFrame
from dolomite_base.read_object import read_object_registry
from multiassayexperiment import MultiAssayExperiment

read_object_registry[
    "multi_sample_dataset"
] = "dolomite_mae.read_multi_assay_experiment"


def read_multi_assay_experiment(
    path: str, metadata: dict, **kwargs
) -> MultiAssayExperiment:
    """Load a
    :py:class:`~multiassayexperiment.MultiAssayExperiment.MultiAssayExperiment`
    from its on-disk representation.

    This method should generally not be called directly but instead be invoked by
    :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path:
            Path to the directory containing the object.

        metadata:
            Metadata for the object.

        kwargs:
            Further arguments, ignored.

    Returns:
        A
        :py:class:`~multiassayexperiment.MultiAssayExperiment.MultiAssayExperiment`
        with file-backed arrays in the assays.
    """

    _sample_path = os.path.join(path, "sample_data")
    _sample_data = None
    if os.path.exists(_sample_path):
        _sample_data = dl.read_object(_sample_path)

    if _sample_data is None:
        raise RuntimeError("Cannot read 'sample_data'.")

    _srow_names = _sample_data.get_row_names()
    if _srow_names is None:
        raise RuntimeError("'sample_data' does not contain 'row_names'.")

    _expts_path = os.path.join(path, "experiments")
    _expts = {}
    _expt_names = []
    _sample_map_data = None
    if os.path.exists(_expts_path):
        with open(os.path.join(_expts_path, "names.json"), "r") as handle:
            _expt_names = json.load(handle)

        if len(_expt_names) > 0:
            _sample_map_path = os.path.join(path, "sample_map.h5")
            _shandle = h5py.File(_sample_map_path, "r")
            _sghandle = _shandle["multi_sample_dataset"]
            _primary = []
            _assay = []
            _colname = []

            for _aidx, _aname in enumerate(_expt_names):
                _expt_read_path = os.path.join(_expts_path, str(_aidx))

                try:
                    _expts[_aname] = dl.read_object(_expt_read_path)
                except Exception as ex:
                    raise RuntimeError(
                        f"failed to load experiment '{_aname}' from '{path}'; "
                        + str(ex)
                    )

                _expt_map = dl.load_vector_from_hdf5(
                    _sghandle[str(_aidx)], expected_type=int, report_1darray=True
                )

                _assay.extend([_aname] * _expts[_aname].shape[1])
                _colname.extend(_expts[_aname].get_column_names())
                _primary.extend([_srow_names[i] for i in _expt_map])

            _sample_map_data = BiocFrame(
                {"primary": _primary, "colname": _colname, "assay": _assay}
            )

    mae = MultiAssayExperiment(
        experiments=_expts, column_data=_sample_data, sample_map=_sample_map_data
    )

    _meta_path = os.path.join(path, "other_data")
    if os.path.exists(_meta_path):
        _meta = dl.read_object(_meta_path)
        mae = mae.set_metadata(_meta.as_dict())

    return mae
